@@ -4,9 +4,11 @@ import platform
 import sys
 import threading
 import time
+import types
 import warnings
-from contextlib import contextmanager
+from contextlib import contextmanager, ExitStack
 from math import inf
+from textwrap import dedent
 
 import attr
 import outcome
@@ -14,9 +16,10 @@ import sniffio
 import pytest
 from async_generator import async_generator
 
-from .tutil import check_sequence_matches, gc_collect_harder
+from .tutil import slow, check_sequence_matches, gc_collect_harder
 from ... import _core
-from ..._timeouts import sleep
+from ..._threads import to_thread_run_sync
+from ..._timeouts import sleep, fail_after
 from ..._util import aiter_compat
 from ...testing import (
     wait_all_tasks_blocked,
@@ -188,8 +191,8 @@ def test_main_and_task_both_crash():
     with pytest.raises(_core.MultiError) as excinfo:
         _core.run(main)
     print(excinfo.value)
-    assert set(type(exc)
-               for exc in excinfo.value.exceptions) == {ValueError, KeyError}
+    assert {type(exc)
+            for exc in excinfo.value.exceptions} == {ValueError, KeyError}
 
 
 def test_two_child_crashes():
@@ -203,8 +206,8 @@ def test_two_child_crashes():
 
     with pytest.raises(_core.MultiError) as excinfo:
         _core.run(main)
-    assert set(type(exc)
-               for exc in excinfo.value.exceptions) == {ValueError, KeyError}
+    assert {type(exc)
+            for exc in excinfo.value.exceptions} == {ValueError, KeyError}
 
 
 async def test_child_crash_wakes_parent():
@@ -335,7 +338,7 @@ async def test_current_statistics(mock_clock):
     await _core.checkpoint()
     await _core.checkpoint()
 
-    with _core.open_cancel_scope(deadline=_core.current_time() + 5):
+    with _core.CancelScope(deadline=_core.current_time() + 5):
         stats = _core.current_statistics()
         print(stats)
         assert stats.seconds_to_next_deadline == 5
@@ -344,9 +347,9 @@ async def test_current_statistics(mock_clock):
     assert stats.seconds_to_next_deadline == inf
 
 
-@attr.s(cmp=False, hash=False)
+@attr.s(eq=False, hash=False)
 class TaskRecorder:
-    record = attr.ib(default=attr.Factory(list))
+    record = attr.ib(factory=list)
 
     def before_run(self):
         self.record.append(("before_run",))
@@ -405,12 +408,13 @@ def test_instruments(recwarn):
 
     _core.run(main, instruments=[r1, r2])
 
-    # It sleeps 5 times, so it runs 6 times
+    # It sleeps 5 times, so it runs 6 times.  Note that checkpoint()
+    # reschedules the task immediately upon yielding, before the
+    # after_task_step event fires.
     expected = (
-        [("before_run",)] +
-        6 * [("schedule", task),
-             ("before", task),
-             ("after", task)] + [("after_run",)]
+        [("before_run",), ("schedule", task)] +
+        [("before", task), ("schedule", task), ("after", task)] * 5 +
+        [("before", task), ("after", task), ("after_run",)]
     )
     assert len(r1.record) > len(r2.record) > len(r3.record)
     assert r1.record == r2.record + r3.record
@@ -442,15 +446,15 @@ def test_instruments_interleave():
         ("schedule", tasks["t2"]),
         {
             ("before", tasks["t1"]),
+            ("schedule", tasks["t1"]),
             ("after", tasks["t1"]),
             ("before", tasks["t2"]),
+            ("schedule", tasks["t2"]),
             ("after", tasks["t2"])
         },
         {
-            ("schedule", tasks["t1"]),
             ("before", tasks["t1"]),
             ("after", tasks["t1"]),
-            ("schedule", tasks["t2"]),
             ("before", tasks["t2"]),
             ("after", tasks["t2"])
         },
@@ -539,15 +543,25 @@ def test_instruments_crash(caplog):
     assert "Instrument has been disabled" in caplog.records[0].message
 
 
-async def test_cancel_scope_repr():
-    # Trivial smoke test
-    with _core.open_cancel_scope() as scope:
-        assert repr(scope).startswith("<cancel scope object")
+async def test_cancel_scope_repr(mock_clock):
+    scope = _core.CancelScope()
+    assert "unbound" in repr(scope)
+    with scope:
+        assert "active" in repr(scope)
+        scope.deadline = _core.current_time() - 1
+        assert "deadline is 1.00 seconds ago" in repr(scope)
+        scope.deadline = _core.current_time() + 10
+        assert "deadline is 10.00 seconds from now" in repr(scope)
+        # when not in async context, can't get the current time
+        assert "deadline" not in await to_thread_run_sync(repr, scope)
+        scope.cancel()
+        assert "cancelled" in repr(scope)
+    assert "exited" in repr(scope)
 
 
 def test_cancel_points():
     async def main1():
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             await _core.checkpoint_if_cancelled()
             scope.cancel()
             with pytest.raises(_core.Cancelled):
@@ -556,7 +570,7 @@ def test_cancel_points():
     _core.run(main1)
 
     async def main2():
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             await _core.checkpoint()
             scope.cancel()
             with pytest.raises(_core.Cancelled):
@@ -565,7 +579,7 @@ def test_cancel_points():
     _core.run(main2)
 
     async def main3():
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             scope.cancel()
             with pytest.raises(_core.Cancelled):
                 await sleep_forever()
@@ -573,7 +587,7 @@ def test_cancel_points():
     _core.run(main3)
 
     async def main4():
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             scope.cancel()
             await _core.cancel_shielded_checkpoint()
             await _core.cancel_shielded_checkpoint()
@@ -584,7 +598,7 @@ def test_cancel_points():
 
 
 async def test_cancel_edge_cases():
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         # Two cancels in a row -- idempotent
         scope.cancel()
         scope.cancel()
@@ -592,7 +606,7 @@ async def test_cancel_edge_cases():
     assert scope.cancel_called
     assert scope.cancelled_caught
 
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         # Check level-triggering
         scope.cancel()
         with pytest.raises(_core.Cancelled):
@@ -606,14 +620,14 @@ async def test_cancel_scope_multierror_filtering():
         raise KeyError
 
     try:
-        with _core.open_cancel_scope() as outer:
+        with _core.CancelScope() as outer:
             try:
                 async with _core.open_nursery() as nursery:
                     # Two children that get cancelled by the nursery scope
                     nursery.start_soon(sleep_forever)  # t1
                     nursery.start_soon(sleep_forever)  # t2
                     nursery.cancel_scope.cancel()
-                    with _core.open_cancel_scope(shield=True):
+                    with _core.CancelScope(shield=True):
                         await wait_all_tasks_blocked()
                     # One child that gets cancelled by the outer scope
                     nursery.start_soon(sleep_forever)  # t3
@@ -622,21 +636,21 @@ async def test_cancel_scope_multierror_filtering():
                     nursery.start_soon(crasher)  # t4
                 # and then our __aexit__ also receives an outer Cancelled
             except _core.MultiError as multi_exc:
-                # This is outside the nursery scope but inside the outer
-                # scope, so the nursery should have absorbed t1 and t2's
-                # exceptions but t3 and t4 should remain, plus the Cancelled
-                # from 'outer'
-                assert len(multi_exc.exceptions) == 3
+                # Since the outer scope became cancelled before the
+                # nursery block exited, all cancellations inside the
+                # nursery block continue propagating to reach the
+                # outer scope.
+                assert len(multi_exc.exceptions) == 5
                 summary = {}
                 for exc in multi_exc.exceptions:
                     summary.setdefault(type(exc), 0)
                     summary[type(exc)] += 1
-                assert summary == {_core.Cancelled: 2, KeyError: 1}
+                assert summary == {_core.Cancelled: 4, KeyError: 1}
                 raise
     except AssertionError:  # pragma: no cover
         raise
     except BaseException as exc:
-        # This is ouside the outer scope, so the two outer Cancelled
+        # This is ouside the outer scope, so all the Cancelled
         # exceptions should have been absorbed, leaving just a regular
         # KeyError from crasher()
         assert type(exc) is KeyError
@@ -661,8 +675,8 @@ async def test_precancelled_task():
 
 
 async def test_cancel_shielding():
-    with _core.open_cancel_scope() as outer:
-        with _core.open_cancel_scope() as inner:
+    with _core.CancelScope() as outer:
+        with _core.CancelScope() as inner:
             await _core.checkpoint()
             outer.cancel()
             with pytest.raises(_core.Cancelled):
@@ -678,7 +692,7 @@ async def test_cancel_shielding():
             # shield protects us from 'outer'
             await _core.checkpoint()
 
-            with _core.open_cancel_scope() as innerest:
+            with _core.CancelScope() as innerest:
                 innerest.cancel()
                 # but it doesn't protect us from scope inside inner
                 with pytest.raises(_core.Cancelled):
@@ -724,7 +738,7 @@ async def test_cancel_inheritance():
 
 
 async def test_cancel_shield_abort():
-    with _core.open_cancel_scope() as outer:
+    with _core.CancelScope() as outer:
         async with _core.open_nursery() as nursery:
             outer.cancel()
             nursery.cancel_scope.shield = True
@@ -749,14 +763,14 @@ async def test_cancel_shield_abort():
             # this worked by cancelling the nursery scope. (When originally
             # written, without these last few lines, the test spuriously
             # passed, even though shield assignment was buggy.)
-            with _core.open_cancel_scope(shield=True):
+            with _core.CancelScope(shield=True):
                 await wait_all_tasks_blocked()
                 assert record == ["sleeping", "cancelled"]
 
 
 async def test_basic_timeout(mock_clock):
     start = _core.current_time()
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         assert scope.deadline == inf
         scope.deadline = start + 1
         assert scope.deadline == start + 1
@@ -768,7 +782,7 @@ async def test_basic_timeout(mock_clock):
     assert not scope.cancel_called
 
     start = _core.current_time()
-    with _core.open_cancel_scope(deadline=start + 1) as scope:
+    with _core.CancelScope(deadline=start + 1) as scope:
         mock_clock.jump(2)
         await sleep_forever()
     # But then the scope swallowed the exception... but we can still see it
@@ -778,7 +792,7 @@ async def test_basic_timeout(mock_clock):
 
     # changing deadline
     start = _core.current_time()
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         await _core.checkpoint()
         scope.deadline = start + 10
         await _core.checkpoint()
@@ -791,11 +805,18 @@ async def test_basic_timeout(mock_clock):
             await _core.checkpoint()
 
 
+@pytest.mark.filterwarnings(
+    "ignore:.*trio.open_cancel_scope:trio.TrioDeprecationWarning"
+)
+async def test_cancel_scope_deprecated(recwarn):
+    assert isinstance(_core.open_cancel_scope(), _core.CancelScope)
+
+
 async def test_cancel_scope_nesting():
     # Nested scopes: if two triggering at once, the outer one wins
-    with _core.open_cancel_scope() as scope1:
-        with _core.open_cancel_scope() as scope2:
-            with _core.open_cancel_scope() as scope3:
+    with _core.CancelScope() as scope1:
+        with _core.CancelScope() as scope2:
+            with _core.CancelScope() as scope3:
                 scope3.cancel()
                 scope2.cancel()
                 await sleep_forever()
@@ -807,8 +828,8 @@ async def test_cancel_scope_nesting():
     assert not scope1.cancelled_caught
 
     # shielding
-    with _core.open_cancel_scope() as scope1:
-        with _core.open_cancel_scope() as scope2:
+    with _core.CancelScope() as scope1:
+        with _core.CancelScope() as scope2:
             scope1.cancel()
             with pytest.raises(_core.Cancelled):
                 await _core.checkpoint()
@@ -822,20 +843,177 @@ async def test_cancel_scope_nesting():
 
     # if a scope is pending, but then gets popped off the stack, then it
     # isn't delivered
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         scope.cancel()
         await _core.cancel_shielded_checkpoint()
     await _core.checkpoint()
     assert not scope.cancelled_caught
 
 
+# Regression test for https://github.com/python-trio/trio/issues/1175
+async def test_unshield_while_cancel_propagating():
+    with _core.CancelScope() as outer:
+        with _core.CancelScope() as inner:
+            outer.cancel()
+            try:
+                await _core.checkpoint()
+            finally:
+                inner.shield = True
+    assert outer.cancelled_caught and not inner.cancelled_caught
+
+
+async def test_cancel_unbound():
+    async def sleep_until_cancelled(scope):
+        with scope, fail_after(1):
+            await sleep_forever()
+
+    # Cancel before entry
+    scope = _core.CancelScope()
+    scope.cancel()
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(sleep_until_cancelled, scope)
+
+    # Cancel after entry
+    scope = _core.CancelScope()
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(sleep_until_cancelled, scope)
+        await wait_all_tasks_blocked()
+        scope.cancel()
+
+    # Shield before entry
+    scope = _core.CancelScope()
+    scope.shield = True
+    with _core.CancelScope() as outer, scope:
+        outer.cancel()
+        await _core.checkpoint()
+        scope.shield = False
+        with pytest.raises(_core.Cancelled):
+            await _core.checkpoint()
+
+    # Can't reuse
+    with _core.CancelScope() as scope:
+        await _core.checkpoint()
+    scope.cancel()
+    await _core.checkpoint()
+    assert scope.cancel_called
+    assert not scope.cancelled_caught
+    with pytest.raises(RuntimeError) as exc_info:
+        with scope:
+            pass  # pragma: no cover
+    assert "single 'with' block" in str(exc_info.value)
+
+    # Can't reenter
+    with _core.CancelScope() as scope:
+        with pytest.raises(RuntimeError) as exc_info:
+            with scope:
+                pass  # pragma: no cover
+        assert "single 'with' block" in str(exc_info.value)
+
+    # Can't enter from multiple tasks simultaneously
+    scope = _core.CancelScope()
+
+    async def enter_scope():
+        with scope:
+            await sleep_forever()
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(enter_scope, name="this one")
+        await wait_all_tasks_blocked()
+
+        with pytest.raises(RuntimeError) as exc_info:
+            with scope:
+                pass  # pragma: no cover
+        assert "single 'with' block" in str(exc_info.value)
+        nursery.cancel_scope.cancel()
+
+    # If not yet entered, cancel_called is true when the deadline has passed
+    # even if cancel() hasn't been called yet
+    scope = _core.CancelScope(deadline=_core.current_time() + 1)
+    assert not scope.cancel_called
+    scope.deadline -= 1
+    assert scope.cancel_called
+    scope.deadline += 1
+    assert scope.cancel_called  # never become un-cancelled
+
+
+async def test_cancel_scope_misnesting():
+    outer = _core.CancelScope()
+    inner = _core.CancelScope()
+    with ExitStack() as stack:
+        stack.enter_context(outer)
+        with inner:
+            with pytest.raises(RuntimeError, match="still within its child"):
+                stack.close()
+        # No further error is raised when exiting the inner context
+
+    # If there are other tasks inside the abandoned part of the cancel tree,
+    # they get cancelled when the misnesting is detected
+    async def task1():
+        with pytest.raises(_core.Cancelled):
+            await sleep_forever()
+
+    # Even if inside another cancel scope
+    async def task2():
+        with _core.CancelScope():
+            with pytest.raises(_core.Cancelled):
+                await sleep_forever()
+
+    with ExitStack() as stack:
+        stack.enter_context(_core.CancelScope())
+        async with _core.open_nursery() as nursery:
+            nursery.start_soon(task1)
+            nursery.start_soon(task2)
+            await wait_all_tasks_blocked()
+            with pytest.raises(RuntimeError, match="still within its child"):
+                stack.close()
+
+    # Variant that makes the child tasks direct children of the scope
+    # that noticed the misnesting:
+    nursery_mgr = _core.open_nursery()
+    nursery = await nursery_mgr.__aenter__()
+    try:
+        nursery.start_soon(task1)
+        nursery.start_soon(task2)
+        nursery.start_soon(sleep_forever)
+        await wait_all_tasks_blocked()
+        nursery.cancel_scope.__exit__(None, None, None)
+    finally:
+        with pytest.raises(RuntimeError) as exc_info:
+            await nursery_mgr.__aexit__(*sys.exc_info())
+        assert "which had already been exited" in str(exc_info.value)
+        assert type(exc_info.value.__context__) is _core.MultiError
+        assert len(exc_info.value.__context__.exceptions) == 3
+        cancelled_in_context = False
+        for exc in exc_info.value.__context__.exceptions:
+            assert isinstance(exc, RuntimeError)
+            assert "closed before the task exited" in str(exc)
+            cancelled_in_context |= isinstance(
+                exc.__context__, _core.Cancelled
+            )
+        assert cancelled_in_context  # for the sleep_forever
+
+    # Trying to exit a cancel scope from an unrelated task raises an error
+    # without affecting any state
+    async def task3(task_status):
+        with _core.CancelScope() as scope:
+            task_status.started(scope)
+            await sleep_forever()
+
+    async with _core.open_nursery() as nursery:
+        scope = await nursery.start(task3)
+        with pytest.raises(RuntimeError, match="from unrelated"):
+            scope.__exit__(None, None, None)
+        scope.cancel()
+
+
+@slow
 async def test_timekeeping():
     # probably a good idea to use a real clock for *one* test anyway...
-    TARGET = 0.1
+    TARGET = 1.0
     # give it a few tries in case of random CI server flakiness
     for _ in range(4):
         real_start = time.perf_counter()
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             scope.deadline = _core.current_time() + TARGET
             await sleep_forever()
         real_duration = time.perf_counter() - real_start
@@ -856,7 +1034,7 @@ async def test_failed_abort():
 
     async def stubborn_sleeper():
         stubborn_task[0] = _core.current_task()
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             stubborn_scope[0] = scope
             record.append("sleep")
             x = await _core.wait_task_rescheduled(lambda _: _core.Abort.FAILED)
@@ -892,7 +1070,7 @@ def test_broken_abort():
         # avoid the warning.
         await _core.checkpoint()
         await _core.checkpoint()
-        with _core.open_cancel_scope() as scope:
+        with _core.CancelScope() as scope:
             scope.cancel()
             # None is not a legal return value here
             await _core.wait_task_rescheduled(lambda _: None)
@@ -1019,7 +1197,7 @@ def test_system_task_crash_KeyboardInterrupt():
 # 5) ...but it's on the run queue, so the timeout is queued to be delivered
 #    the next time that it's blocked.
 async def test_yield_briefly_checks_for_timeout(mock_clock):
-    with _core.open_cancel_scope(deadline=_core.current_time() + 5):
+    with _core.CancelScope(deadline=_core.current_time() + 5):
         await _core.checkpoint()
         with pytest.raises(_core.Cancelled):
             mock_clock.jump(10)
@@ -1028,7 +1206,7 @@ async def test_yield_briefly_checks_for_timeout(mock_clock):
 
 # This tests that sys.exc_info is properly saved/restored as we swap between
 # tasks. It turns out that the interpreter automagically handles this for us
-# so there's no special code in trio required to pass this test, but it's
+# so there's no special code in Trio required to pass this test, but it's
 # still nice to know that it works :-).
 #
 # Update: it turns out I was right to be nervous! see the next test...
@@ -1318,7 +1496,7 @@ def test_TrioToken_run_sync_soon_starvation_resistance():
     assert len(record) == 2
     assert record[0] == "starting"
     assert record[1][0] == "run finished"
-    assert record[1][1] >= 20
+    assert record[1][1] >= 19
 
 
 def test_TrioToken_run_sync_soon_threaded_stress_test():
@@ -1372,7 +1550,7 @@ async def test_TrioToken_run_sync_soon_massive_queue():
 
 
 async def test_slow_abort_basic():
-    with _core.open_cancel_scope() as scope:
+    with _core.CancelScope() as scope:
         scope.cancel()
         with pytest.raises(_core.Cancelled):
             task = _core.current_task()
@@ -1407,8 +1585,8 @@ async def test_slow_abort_edge_cases():
         await _core.checkpoint()
         record.append("done")
 
-    with _core.open_cancel_scope() as outer1:
-        with _core.open_cancel_scope() as outer2:
+    with _core.CancelScope() as outer1:
+        with _core.CancelScope() as outer2:
             async with _core.open_nursery() as nursery:
                 # So we have a task blocked on an operation that can't be
                 # aborted immediately
@@ -1516,8 +1694,8 @@ async def test_spawn_name():
 async def test_current_effective_deadline(mock_clock):
     assert _core.current_effective_deadline() == inf
 
-    with _core.open_cancel_scope(deadline=5) as scope1:
-        with _core.open_cancel_scope(deadline=10) as scope2:
+    with _core.CancelScope(deadline=5) as scope1:
+        with _core.CancelScope(deadline=10) as scope2:
             assert _core.current_effective_deadline() == 5
             scope2.deadline = 3
             assert _core.current_effective_deadline() == 3
@@ -1535,6 +1713,8 @@ async def test_current_effective_deadline(mock_clock):
     assert _core.current_effective_deadline() == inf
 
 
+# @coroutine is deprecated since python 3.8, which is fine with us.
+@pytest.mark.filterwarnings("ignore:.*@coroutine.*:DeprecationWarning")
 def test_nice_error_on_bad_calls_to_run_or_spawn():
     def bad_call_run(*args):
         _core.run(*args)
@@ -1573,10 +1753,6 @@ def test_nice_error_on_bad_calls_to_run_or_spawn():
                 bad_call(asyncio.Future())
             assert "asyncio" in str(excinfo.value)
 
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(generator_based_coro)
-            assert "asyncio" in str(excinfo.value)
-
             with pytest.raises(TypeError) as excinfo:
                 bad_call(lambda: asyncio.Future())
             assert "asyncio" in str(excinfo.value)
@@ -1607,14 +1783,31 @@ def test_nice_error_on_bad_calls_to_run_or_spawn():
 
 
 def test_calling_asyncio_function_gives_nice_error():
-    async def misguided():
+    async def child_xyzzy():
         import asyncio
         await asyncio.Future()
 
+    async def misguided():
+        await child_xyzzy()
+
     with pytest.raises(TypeError) as excinfo:
         _core.run(misguided)
 
     assert "asyncio" in str(excinfo.value)
+    # The traceback should point to the location of the foreign await
+    assert any(  # pragma: no branch
+        entry.name == "child_xyzzy" for entry in excinfo.traceback
+    )
+
+
+async def test_asyncio_function_inside_nursery_does_not_explode():
+    # Regression test for https://github.com/python-trio/trio/issues/552
+    with pytest.raises(TypeError) as excinfo:
+        async with _core.open_nursery() as nursery:
+            import asyncio
+            nursery.start_soon(sleep_forever)
+            await asyncio.Future()
+    assert "asyncio" in str(excinfo.value)
 
 
 async def test_trivial_yields():
@@ -1629,15 +1822,16 @@ async def test_trivial_yields():
         async with _core.open_nursery():
             pass
 
-    with _core.open_cancel_scope() as cancel_scope:
+    with _core.CancelScope() as cancel_scope:
         cancel_scope.cancel()
         with pytest.raises(_core.MultiError) as excinfo:
             async with _core.open_nursery():
                 raise KeyError
         assert len(excinfo.value.exceptions) == 2
-        assert set(type(e) for e in excinfo.value.exceptions) == {
-            KeyError, _core.Cancelled
-        }
+        assert {type(e)
+                for e in excinfo.value.exceptions} == {
+                    KeyError, _core.Cancelled
+                }
 
 
 async def test_nursery_start(autojump_clock):
@@ -1708,7 +1902,7 @@ async def test_nursery_start(autojump_clock):
         task_status.started("hi")
 
     async with _core.open_nursery() as nursery:
-        with _core.open_cancel_scope() as cs:
+        with _core.CancelScope() as cs:
             cs.cancel()
             with pytest.raises(_core.Cancelled):
                 await nursery.start(just_started)
@@ -1722,13 +1916,12 @@ async def test_nursery_start(autojump_clock):
         raise KeyError("whoopsiedaisy")
 
     async with _core.open_nursery() as nursery:
-        with _core.open_cancel_scope() as cs:
+        with _core.CancelScope() as cs:
             cs.cancel()
             with pytest.raises(_core.MultiError) as excinfo:
                 await nursery.start(raise_keyerror_after_started)
-    assert set(type(e) for e in excinfo.value.exceptions) == {
-        _core.Cancelled, KeyError
-    }
+    assert {type(e)
+            for e in excinfo.value.exceptions} == {_core.Cancelled, KeyError}
 
     # trying to start in a closed nursery raises an error immediately
     async with _core.open_nursery() as closed_nursery:
@@ -1843,7 +2036,7 @@ async def test_nursery_stop_iteration():
 
 
 async def test_nursery_stop_async_iteration():
-    class it(object):
+    class it:
         def __init__(self, count):
             self.count = count
             self.val = 0
@@ -1856,7 +2049,7 @@ async def test_nursery_stop_async_iteration():
             self.val += 1
             return val
 
-    class async_zip(object):
+    class async_zip:
         def __init__(self, *largs):
             self.nexts = [obj.__anext__ for obj in largs]
 
@@ -1980,9 +2173,33 @@ def test_system_task_contexts():
     _core.run(inner)
 
 
+def test_Nursery_init():
+    check_Nursery_error = pytest.raises(
+        TypeError, match='no public constructor available'
+    )
+
+    with check_Nursery_error:
+        _core._run.Nursery(None, None)
+
+
+async def test_Nursery_private_init():
+    # context manager creation should not raise
+    async with _core.open_nursery() as nursery:
+        assert False == nursery._closed
+
+
+def test_Nursery_subclass():
+    with pytest.raises(
+        TypeError, match='`Nursery` does not support subclassing'
+    ):
+
+        class Subclass(_core._run.Nursery):
+            pass
+
+
 def test_Cancelled_init():
     check_Cancelled_error = pytest.raises(
-        RuntimeError, match='should not be raised directly'
+        TypeError, match='no public constructor available'
     )
 
     with check_Cancelled_error:
@@ -1992,7 +2209,30 @@ def test_Cancelled_init():
         _core.Cancelled()
 
     # private constructor should not raise
-    _core.Cancelled._init()
+    _core.Cancelled._create()
+
+
+def test_Cancelled_str():
+    cancelled = _core.Cancelled._create()
+    assert str(cancelled) == 'Cancelled'
+
+
+def test_Cancelled_subclass():
+    with pytest.raises(
+        TypeError, match='`Cancelled` does not support subclassing'
+    ):
+
+        class Subclass(_core.Cancelled):
+            pass
+
+
+def test_CancelScope_subclass():
+    with pytest.raises(
+        TypeError, match='`CancelScope` does not support subclassing'
+    ):
+
+        class Subclass(_core.CancelScope):
+            pass
 
 
 def test_sniffio_integration():
@@ -2015,3 +2255,185 @@ async def test_Task_custom_sleep_data():
     assert task.custom_sleep_data == 1
     await _core.checkpoint()
     assert task.custom_sleep_data is None
+
+
+@types.coroutine
+def async_yield(value):
+    yield value
+
+
+async def test_permanently_detach_coroutine_object():
+    task = None
+    pdco_outcome = None
+
+    async def detachable_coroutine(task_outcome, yield_value):
+        await sleep(0)
+        nonlocal task, pdco_outcome
+        task = _core.current_task()
+        pdco_outcome = await outcome.acapture(
+            _core.permanently_detach_coroutine_object, task_outcome
+        )
+        await async_yield(yield_value)
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(
+            detachable_coroutine, outcome.Value(None), "I'm free!"
+        )
+
+    # If we get here then Trio thinks the task has exited... but the coroutine
+    # is still iterable
+    assert pdco_outcome is None
+    assert task.coro.send("be free!") == "I'm free!"
+    assert pdco_outcome == outcome.Value("be free!")
+    with pytest.raises(StopIteration):
+        task.coro.send(None)
+
+    # Check the exception paths too
+    task = None
+    pdco_outcome = None
+    with pytest.raises(KeyError):
+        async with _core.open_nursery() as nursery:
+            nursery.start_soon(
+                detachable_coroutine, outcome.Error(KeyError()), "uh oh"
+            )
+    throw_in = ValueError()
+    assert task.coro.throw(throw_in) == "uh oh"
+    assert pdco_outcome == outcome.Error(throw_in)
+    with pytest.raises(StopIteration):
+        task.coro.send(None)
+
+    async def bad_detach():
+        async with _core.open_nursery():
+            with pytest.raises(RuntimeError) as excinfo:
+                await _core.permanently_detach_coroutine_object(
+                    outcome.Value(None)
+                )
+            assert "open nurser" in str(excinfo.value)
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(bad_detach)
+
+
+async def test_detach_and_reattach_coroutine_object():
+    unrelated_task = None
+    task = None
+
+    async def unrelated_coroutine():
+        nonlocal unrelated_task
+        unrelated_task = _core.current_task()
+
+    async def reattachable_coroutine():
+        await sleep(0)
+
+        nonlocal task
+        task = _core.current_task()
+
+        def abort_fn(_):  # pragma: no cover
+            return _core.Abort.FAILED
+
+        got = await _core.temporarily_detach_coroutine_object(abort_fn)
+        assert got == "not trio!"
+
+        await async_yield(1)
+        await async_yield(2)
+
+        with pytest.raises(RuntimeError) as excinfo:
+            await _core.reattach_detached_coroutine_object(
+                unrelated_task, None
+            )
+        assert "does not match" in str(excinfo.value)
+
+        await _core.reattach_detached_coroutine_object(task, "byebye")
+
+        await sleep(0)
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(unrelated_coroutine)
+        nursery.start_soon(reattachable_coroutine)
+        await wait_all_tasks_blocked()
+        assert unrelated_task is not None
+        assert task is not None
+
+        # Okay, it's detached. Here's our coroutine runner:
+        assert task.coro.send("not trio!") == 1
+        assert task.coro.send(None) == 2
+        assert task.coro.send(None) == "byebye"
+
+        # Now it's been reattached, and we can leave the nursery
+
+
+async def test_detached_coroutine_cancellation():
+    abort_fn_called = False
+    task = None
+
+    async def reattachable_coroutine():
+        await sleep(0)
+
+        nonlocal task
+        task = _core.current_task()
+
+        def abort_fn(_):
+            nonlocal abort_fn_called
+            abort_fn_called = True
+            return _core.Abort.FAILED
+
+        await _core.temporarily_detach_coroutine_object(abort_fn)
+        await _core.reattach_detached_coroutine_object(task, None)
+        with pytest.raises(_core.Cancelled):
+            await sleep(0)
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(reattachable_coroutine)
+        await wait_all_tasks_blocked()
+        assert task is not None
+        nursery.cancel_scope.cancel()
+        task.coro.send(None)
+
+    assert abort_fn_called
+
+
+def test_async_function_implemented_in_C():
+    # These used to crash because we'd try to mutate the coroutine object's
+    # cr_frame, but C functions don't have Python frames.
+
+    ns = {"_core": _core}
+    try:
+        exec(
+            dedent(
+                """
+                async def agen_fn(record):
+                    assert not _core.currently_ki_protected()
+                    record.append("the generator ran")
+                    yield
+                """
+            ),
+            ns,
+        )
+    except SyntaxError:
+        pytest.skip("Requires Python 3.6+")
+    else:
+        agen_fn = ns["agen_fn"]
+
+    run_record = []
+    agen = agen_fn(run_record)
+    _core.run(agen.__anext__)
+    assert run_record == ["the generator ran"]
+
+    async def main():
+        start_soon_record = []
+        agen = agen_fn(start_soon_record)
+        async with _core.open_nursery() as nursery:
+            nursery.start_soon(agen.__anext__)
+        assert start_soon_record == ["the generator ran"]
+
+    _core.run(main)
+
+
+async def test_very_deep_cancel_scope_nesting():
+    # This used to crash with a RecursionError in CancelStatus.recalculate
+    with ExitStack() as exit_stack:
+        outermost_scope = _core.CancelScope()
+        exit_stack.enter_context(outermost_scope)
+        for _ in range(5000):
+            exit_stack.enter_context(_core.CancelScope())
+        outermost_scope.cancel()
